'''
Created on Jan 31, 2010

@author: richard
'''

from random import randint, random

class Breeder(object):
    
    def __init__(self, mutation_rate=0.01, shorten_rate=0.1, lengthen_rate=0.1):
        self.mutation_rate = mutation_rate
        self.shorten_rate = shorten_rate
        self.lengthen_rate = lengthen_rate
        self.selection = ['1','2','3','4','5','6','7','8','9','a','b','c','d','e','f','g']
    
    def breed(self, mum, dad):
        pass
    
    def shorten(self, gene):
        if len(gene) == 1:
            return gene
        
        if random() > self.shorten_rate:
            return gene
        
        p = randint(0, len(gene) - 1)
        gene = gene[:p] + gene[p + 1:]
        return gene
    
    def lengthen(self, gene):
        if random() > self.lengthen_rate:
            return gene
        
        p = randint(0, len(gene) - 1)
        s = randint(0, len(self.selection) - 1)
        gene = gene[:p] + self.selection[s] + gene[p:]
        return gene
    
    def mutate(self, gene):
        ngene = ''
        for b in gene:
            if random() < self.mutation_rate:
                s = randint(0, len(self.selection) - 1)
                ngene += self.selection[s]
            else:
                ngene += b
        return ngene

"""
Crossovers can be performed by sequentially copying genes 
from one parent, but with some frequency the source genotype 
is switched to the other parent. This causes adjacent 
genes to be more likely to stick together than genes at opposite 
ends of the sequence. Each pair of genes has a linkage 
probability depending on their distance from each other.
"""
class Crossover(Breeder):
    def breed(self, mum, dad):
        pass



"""
Each gene can be independently copied from one parent or the 
other with equal probability. If the parent genes each 
correspond to a point in N-dimensional genetic space, then 
the genes of the possible children using this method correspond 
to the 2N corners of the N-dimensional rectangular solid 
connecting the two parent points. This method is the most 
commonly used in this work.
"""
class Independent(Breeder):
    
    def __init__(self, mutation_rate=0.01, shorten_rate=0.1, lengthen_rate=0.01):
        Breeder.__init__(self, mutation_rate, shorten_rate, lengthen_rate)
        
    def breed(self, mum, dad):
        mgenes = mum.split('0')
        dgenes = dad.split('0')
        parents = [mgenes, dgenes]
        
        ngene = ''
        for i in range(len(mgenes)):
            p = randint(0, 1)
            gene = parents[p][i]
            gene = self.shorten(gene)
            gene = self.lengthen(gene)
            gene = self.mutate(gene)
            ngene += gene + '0'
        return ngene[:-1] 


"""
Each gene can receive a random percentage, p, of one parent's 
genes, and a 1 - p percentage of the other parent's genes. 
If the percentage is the same for each gene, linear interpolation 
between the parent genotypes results, and the children will fall 
randomly on the line between the N-dimensional points of the 
parents. If evenly spaced samples along this line were generated, 
a genetic dissolve could be made that would cause a smooth 
transition between the parent phenotypes if the changing parameters 
had continuous effects on the phenotypes. This is an example of 
utilizing the underlying genetic representation for specific 
manipulation of the results. Interpolation could also be 
performed with three parents to create children that fall on a 
triangular region of a plane in the N-dimensional genetic space. 
"""
class Dissolve(Breeder):
    def breed(self, mum, dad):
        pass


"""
Finally, each new gene can receive a random value between the two 
parent values of that gene. This is like the interpolation scheme 
above, except each gene is independently interpolated between 
the parent genes. This method results in possible children anywhere 
within the N-dimensional rectangular solid connecting the parent points.
"""
class Interpolate(Breeder):
    def breed(self, mum, dad):
        pass

