'''
Created on Jan 31, 2010

@author: richard
'''

from model import Algorithm, Maze, Ant, Game, DAO
from fitness import UnfinishedFitness, FinishedFitness
from breed import Independent

import os

class Settings(object):
    
    def __init__(self):
        self.maze_width = 20
        self.maze_height = 20
        self.maze_entrance = (1,1)
        self.maze_exit = (20,20)
        self.population_size = 500
        self.maze_intersects = 5
        self.finisher_threshold = 10
        self.top_slice = 100
        self.improvement_attempts = 5
        self.maze_count = 10
        self.data_dir = 'data'
        self.maze_file = 'maze.txt'
        self.ants_dir = 'ants'
        self.maze_dir = 'mazes'
        self.save_finishers = False
        self.save_best = False
        self.save_all = False
        self.unfinished_fitness = UnfinishedFitness()
        self.finished_fitness = FinishedFitness()
        self.breeder = Independent()
        self.dao = DAO()

class State(object):
    
    def __init__(self):
        self.ant = None
        self.algorithm = None
        self.fitness = 0.0

def generalise_breadth_first(settings):
    
    # generate the appropriate number of mazes
    mazes = []
    for i in range(settings.maze_count):
        maze = Maze(settings.maze_width, settings.maze_height, settings.maze_entrance, settings.maze_exit)
        maze.random(settings.maze_intersects)
        maze.save(os.path.join(settings.data_dir, settings.maze_dir))
        mazes.append(maze)
    
    population = []
    append_random(population, settings.population_size)
    
    finishers = []
    while True:
        # evaluate each ant against each maze
        for maze in mazes:
            evaluate_population(settings, population, maze, settings.unfinished_fitness, True)
        
        # sort and breed
        population.sort(lambda x, y: 1 if x.fitness < y.fitness else 0 if x.fitness == y.fitness else -1)
        finishers = finishers + [state for state in population if state.fitness == settings.maze_count]
        
        print 'number of finishers: ' + str(len(finishers))
        print 'best candidate: ' + population[0].algorithm.dna + ' - ' + str(population[0].fitness)
        
        if len(finishers) > settings.finisher_threshold:
            break
        
        population = next_generation(settings, population)
    
    # for each one that can finish all the mazes
    for finisher in finishers:
        print 'saving ' + finisher.algorithm.dna
        # for each of the mazes
        for maze in mazes:
            # replay into a new ant
            ant = Ant()
            game = Game(ant, maze, finisher.algorithm)
            game.play()
            
            # save
            settings.dao.save_solution(finisher.algorithm, ant, maze, 'Richard Jones', os.path.join(settings.data_dir, settings.ants_dir))

# this doesn't work so well, use generalise_breadth_first
def generalise_depth_first(settings):
    
    # generate the appropriate number of mazes
    mazes = []
    for i in range(settings.maze_count):
        maze = Maze(settings.maze_width, settings.maze_height, settings.maze_entrance, settings.maze_exit)
        maze.random(settings.maze_intersects)
        maze.save(os.path.join(settings.data_dir, settings.maze_dir))
        mazes.append(maze)
    
    # generalisation is a bit tricky:
    #
    # - evolve a population of ants which can finish one maze
    # - pit them all against subsequent mazes until fewer than some limit succeed
    # - breed the survivors, then start again
    # - terminate once we have an ant which can complete all 10
    
    population = []
    append_random(population, settings.population_size)
    
    while True:
        current_maze = 1
        
        # stage 1: get a population which can pass the first maze
        population = evolve_finishers(settings, mazes[0], population)
        print 'evolved finishers'
        
        while True:
            
            print 'starting generalisation'
            
            # obtain the ants which can finish the current maze
            evaluate_population(settings, population, mazes[current_maze], settings.unfinished_fitness)
            population.sort(lambda x, y: 1 if x.fitness < y.fitness else 0 if x.fitness == y.fitness else -1)
            finishers = [state for state in population if state.fitness == 0.0]
            
            # is it big enough to continue
            if len(finishers) < settings.top_slice:
                print 'not enough finishers: ' + str(len(finishers))
                break
            else:
                population = finishers
            
            mazes[current_maze].render()
            print 'maze ' + str(current_maze) + ':' + str(len(population)) + ' finishers'
            
            # if we get here, we can continue up the chain
            current_maze += 1
            
            # have we completed all the mazes
            if current_maze > len(mazes):
                break
        
        if current_maze > len(mazes):
            break
        
        # breed the population up to the appropriate number of participants
        population = next_generation(settings, population)
    
    # save the output
    for state in population:
        state.ant.save(os.path.join(settings.data_dir, settings.ants_dir, state.algorithm.dna))

def specialise(settings, maze=None):
    if maze is None:
        maze = Maze(settings.maze_width, settings.maze_height, settings.maze_entrance, settings.maze_exit)
        maze.random(settings.maze_intersects)
        maze.save(os.path.join(settings.data_dir, settings.maze_dir))
    
    population = []
    append_random(population, settings.population_size)
    best = 0.0
    no_improvement = 0;
    
    while True:
        population = evolve_finishers(settings, maze, population)
        evaluate_population(settings, population, maze, settings.finished_fitness)
        population.sort(lambda x, y: 1 if x.fitness < y.fitness else 0 if x.fitness == y.fitness else -1)
        
        print 'shortest path: ' + population[0].algorithm.dna + ' - ' + str(population[0].fitness)
        
        if best < population[0].fitness:
            best = population[0].fitness
            no_improvement = 0
        else:
            no_improvement += 1
        
        settings.dao.save_solution(population[0].algorithm, population[0].ant, maze, 'Richard Jones', os.path.join(settings.data_dir, settings.ants_dir))
        #ant_file = os.path.join(settings.data_dir, settings.ants_dir, population[0].algorithm.dna)
        #population[0].ant.save(ant_file)
        
        if no_improvement >= settings.improvement_attempts:
            break
        
        population = next_generation(settings, population)
    
# evolve a number of finishers to a given, or randomly
# generated maze.  This will cause the maze and the
# surviving ants to be saved to file
def evolve_finishers(settings, maze=None, population=None):
    if maze is None:
        maze = Maze(settings.maze_width, settings.maze_height, settings.maze_entrance, settings.maze_exit)
        maze.random(settings.maze_intersects)
        maze.save(os.path.join(settings.data_dir, settings.maze_dir))
    
    if population is None:
        population = []
        append_random(population, settings.population_size)
    
    fins = get_finishers(population, maze, settings)
    return fins

###############################################################################

def append_random(population, count):
    for i in range(count):
        state = State()
        alg = Algorithm()
        alg.random()
        state.algorithm = alg
        population.append(state)

def extract_finishers(evaluation):
    finishers = []
    for dna in evaluation.keys():
        if evaluation[dna] == 1.0:
            finishers.append(dna)
    return finishers

def get_finishers(population, maze, settings):
    f = settings.unfinished_fitness
    finishers = []
    while True:
        evaluate_population(settings, population, maze, f)
        population.sort(lambda x, y: 1 if x.fitness < y.fitness else 0 if x.fitness == y.fitness else -1)
        
        best = None
        best_non = None
        
        for state in population:
            if state.fitness == 1.0:
                finishers.append(state)
                if settings.save_finishers:
                    settings.dao.save_solution(state.algorithm, state.ant, maze, 'Richard Jones', os.path.join(settings.data_dir, settings.ants_dir))
                    # state.ant.save(os.path.join(settings.data_dir, settings.ants_dir, state.algorithm.dna))
                if best is None: 
                    best = state
            else:
                if best is None:
                    if settings.save_best:
                        settings.dao.save_solution(state.algorithm, state.ant, maze, 'Richard Jones', os.path.join(settings.data_dir, settings.ants_dir))
                        # state.ant.save(os.path.join(settings.data_dir, settings.ants_dir, state.algorithm.dna))
                    best = state
                if best_non is None:
                    best_non = state
        
        print 'number of finishers: ' + str(len(finishers))
        print 'best candidate: ' + best.algorithm.dna + ' - ' + str(best.fitness)
        print 'best non finisher: ' + best_non.algorithm.dna + ' - ' + str(best_non.fitness)
    
        if len(finishers) >= settings.finisher_threshold:
            print "finished - threshold reached"
            break
        
        population = next_generation(settings, population)
    
    maze.render()
    return finishers

def next_generation(settings, evaluation):
    multiplicity = (settings.population_size / (settings.top_slice / 2)) - 1
    freespace = settings.population_size - (settings.top_slice / 2)
    population = []
    for i in range(0, settings.top_slice, 2):
        mstate = evaluation[i]
        dstate = evaluation[i + 1]
        for j in range(multiplicity):
            child = settings.breeder.breed(mstate.algorithm.dna, dstate.algorithm.dna)
            cstate = State()
            cstate.algorithm = Algorithm(child)
            population.append(cstate)
    append_random(population, freespace)
    return population

def evaluate_population(settings, population, maze, fitness, additive=False):
    evaluation = {}
    for i in range(len(population)):
        ant = Ant()
        alg = population[i].algorithm
        population[i].ant = ant
        
        game = Game(ant, maze, alg)
        game.play()
        
        if additive:
            population[i].fitness += fitness.score(ant, maze)
        else:
            population[i].fitness = fitness.score(ant, maze)
