'''
Created on Jan 31, 2010

@author: richard
'''

from random import randint
from math import sqrt
import actions
import uuid
import os

class DAO(object):
    
    def save_solution(self, algorithm, ant, maze, credit, dir):
        content = ''
        content += maze.name + '\n'
        content += algorithm.dna + '\n'
        content += credit + '\n'
        content += ant.path_content()
        
        f = open(os.path.join(dir, ant.name), "w")
        f.write(content)

class Ant(object):
    
    def __init__(self):
        self.path = []
        self.endpoints = []
        self.current = ()
        self.start = ()
        self.end = ()
        self.complete = False
        self.name = str(uuid.uuid4())
    
    def pass_through(self, x, y):
        self.path.append((x, y))
        self.current = (x, y)
        if self.current == self.end:
            self.complete = True
    
    def end_on(self, coords):
        self.endpoints.append(coords)
    
    def path_content(self):
        s = ''
        for (x, y) in self.path:
            s += str(x) + ' ' + str(y) + '\n'
        return s
    
    def save(self, file):
        s = ''
        for (x, y) in self.path:
            s += str(x) + ' ' + str(y) + '\n'
        f = open(file, 'w')
        f.write(s)

class Maze(object):
    
    def __init__(self, w=30, h=30, entrance=(1,1), exit=(30, 30)):
        self.width = w
        self.height = h
        self.exit = exit
        self.entrance = entrance
        self.name = str(uuid.uuid4())
        
        # have to build the grid slowly to avoid reference problems
        self.grid = []
        self.grid.append([1 for i in range(h+2)])
        for j in range(h):
            self.grid.append([1] + [0 for i in range(w)] + [1])
        self.grid.append([1 for i in range(h+2)])
    
    def random(self, points=5):
        (sx, sy) = self.entrance
        (ex, ey) = self.exit
        
        # generate a set of points to pass through
        reach = []
        for i in range(points):
            rx = randint(1, self.width)
            ry = randint(1, self.height)
            reach.append((rx, ry))
        
        #print 'reach: ' + str(reach)
        
        clear_path = [(sx, sy)]
        reached = []
        cx = sx
        cy = sy
        for j in range(len(reach) - 1):
            
            # locate the nearest intersection
            nx = None
            ny = None
            dd = None
            for (tx, ty) in reach:
                #print 'inspecting: ' + str((tx, ty))
                #print 'already reached: ' + str(reached)
                if (tx, ty) not in reached:
                    dx = tx - cx
                    dy = ty - cy
                    d = sqrt(dx*dx + dy*dy)
                    if dd is None or d < dd:
                        nx = dx
                        ny = dy
            tx = cx + nx
            ty = cy + ny
            #print 'closest: ' + str((tx, ty))
            
            # plot a line between the intersections
            xstep = 1 if nx > 0 else -1
            ystep = 1 if ny > 0 else -1
            #print 'x loop: ' + str(cx) + ',' + str(tx) + ',' + str(xstep)
            for i in range(cx, tx + xstep, xstep):
                #print 'adding to (x) path: ' + str((i, cy))
                clear_path.append((i, cy))
            #print 'y loop: ' + str(cy) + ',' + str(ty) + ',' + str(ystep)
            for i in range(cy, ty + ystep, ystep):
                #print 'adding to (y) path: ' + str((tx, i))
                clear_path.append((tx, i))
            #print 'clear path: ' + str(clear_path)
            
            # update to the current place
            cx = tx
            cy = ty
            reached.append((tx, ty))
        
        # plot a line between the intersections
        nx = ex - cx
        ny = ey - cy
        xstep = 1 if nx > 0 else -1
        ystep = 1 if ny > 0 else -1
        #print 'x loop: ' + str(cx) + ',' + str(ex) + ',' + str(xstep)
        for i in range(cx, ex + xstep, xstep):
            #print 'adding to (x) path: ' + str((i, cy))
            clear_path.append((i, cy))
        #print 'y loop: ' + str(cy) + ',' + str(ey) + ',' + str(ystep)
        for i in range(cy, ey + ystep, ystep):
            #print 'adding to (y) path: ' + str((ex, i))
            clear_path.append((ex, i))
        #print 'clear path: ' + str(clear_path)
        
        # now fill the remaining area with obstacles
        for x in range(self.width):
            for y in range(self.height):
                #print 'considering: ' + str((x,y))
                if self.is_wall(x, y):
                    #print 'wall'
                    continue
                if (x, y) in clear_path:
                    #print 'to be kept clear'
                    continue
                if (x,y) == self.entrance or (x, y) == self.exit:
                    continue
                v = randint(0, 1)
                #print 'set point ' + str((x, y)) + ' = ' + str(v)
                self.set_point(x, y, v)
        
        #print clear_path
        
    def set_point(self, x, y, v):
        self.grid[y][x] = v
        
    def get_point(self, x, y):
        return self.grid[y][x]
    
    def is_wall(self, x, y):
        return self.grid[y][x] > 0
    
    def render(self, ant=None):
        if ant is not None:
            (x, y) = ant.current
        (sx, sy) = self.entrance
        (ex, ey) = self.exit
        for i in range(len(self.grid) - 1, -1, -1):
            temprow = []
            for j in range(len(self.grid[i])):
                if (i, j) == (sx, sy):
                    temprow.append('s')
                elif (i, j) == (ex, ey):
                    temprow.append('e')
                elif ant is not None and (i, j) == (y, x):
                    temprow.append('a')
                else:
                    if self.grid[i][j] == 1:
                        temprow.append('X')
                    else:
                        temprow.append(' ')
            print "%.2d" % (i) + " --  " + "".join(temprow)
    
    def save(self, dir):
        (sx, sy) = self.entrance
        (ex, ey) = self.exit
        s = str(sx) + ' ' + str(sy) + ' ' + str(ex) + ' ' + str(ey) + '\n'
        for i in range(len(self.grid) - 1, -1, -1):
            for j in range(len(self.grid[i])):
                if self.grid[i][j] == 1:
                    s += '1'
                else:
                    s += '0'
                s += ' '
            s += '\n'
        f = open(os.path.join(dir, self.name), 'w')
        f.write(s)
    
class Game(object):
    
    def __init__(self, ant, maze, alg):
        self.ant = ant
        self.maze = maze
        self.alg = alg
        
        self.ant.end = self.maze.exit
        self.ant.start = self.maze.entrance
        self.ant.current = self.maze.entrance
    
    def play(self):
        
        current = (None, None)
        limit = 10
        count = 0
        while current not in self.ant.endpoints and count < limit:
            self.ant.end_on(self.ant.current)
            
            state = self.get_ant_state()
            ops = self.alg.get_ops(state)
            
            #print state
            #print self.alg.state_description(state)
            #print ''
            
            for op in ops:
                op(self.ant, self.maze)
                if self.ant.complete:
                    break
            
            current = self.ant.current
            count += 1
            
            #self.maze.render(self.ant)
        
        if current in self.ant.endpoints:
            #print 'terminated due to circularity'
            pass
        
        if count == limit:        
            #print 'hard execution limit reached: ' + str(limit)
            pass
        
        if self.ant.complete:
            # print 'exit reached'
            pass
    
    def get_ant_state(self):
        (x, y) = self.ant.current
        
        north_wall = False
        south_wall = False
        east_wall = False
        west_wall = False
        
        if self.maze.is_wall(x, y + 1):
            north_wall = True
        
        if self.maze.is_wall(x, y - 1):
            south_wall = True
        
        if self.maze.is_wall(x + 1, y):
            east_wall = True
        
        if self.maze.is_wall(x - 1, y):
            west_wall = True
        
        # lookup matrix for state
        #        north, east, south, west
        lookup = {
                  (True, True, True, True) : 'encased',
                  (True, True, True, False) : 'dead_end_west',
                  (True, True, False, True) : 'dead_end_south',
                  (True, True, False, False) : 'corner_north_east',
                  (True, False, True, True) : 'dead_end_east',
                  (True, False, True, False) : 'horizontal_tunnel',
                  (True, False, False, True) : 'corner_north_west',
                  (True, False, False, False) : 'wall_north',
                  (False, True, True, True) : 'dead_end_north',
                  (False, True, True, False) : 'corner_south_east',
                  (False, True, False, True) : 'vertical_tunnel',
                  (False, True, False, False) : 'wall_east',
                  (False, False, True, True) : 'corner_south_west',
                  (False, False, True, False) : 'wall_south',
                  (False, False, False, True) : 'wall_west',
                  (False, False, False, False) : 'clear'
                  }
        
        return lookup[(north_wall, east_wall, south_wall, west_wall)]
        
class Algorithm(object):
    
    def __init__(self, dna=None):
        self.states = {'encased' : None, 
                       'dead_end_west' : None, 
                       'dead_end_south' : None,
                       'corner_north_east' : None, 
                       'dead_end_east' : None, 
                       'horizontal_tunnel' : None,
                       'corner_north_west' : None, 
                       'wall_north' : None, 
                       'dead_end_north' : None,
                       'corner_south_east' : None, 
                       'vertical_tunnel' : None,
                       'wall_east' : None,
                       'corner_south_west' : None,
                       'wall_south' :  None,
                       'wall_west' : None,
                       'clear' : None
                       }
        
        self.state_map = {1 : 'encased', 
                          2 : 'dead_end_west', 
                          3 : 'dead_end_south',
                          4 : 'corner_north_east', 
                          5 : 'dead_end_east', 
                          6 : 'horizontal_tunnel',
                          7 : 'corner_north_west', 
                          8 : 'wall_north', 
                          9 : 'dead_end_north',
                          10 : 'corner_south_east', 
                          11 : 'vertical_tunnel',
                          12 : 'wall_east',
                          13 : 'corner_south_west',
                          14 : 'wall_south',
                          15 : 'wall_west',
                          16 : 'clear'
                          }
        
        self.funcs = {'1' : actions.move_left, '2' : actions.move_right, '3' : actions.move_up, '4' : actions.move_down,
             '5' : actions.move_left_until_vertical, '6' : actions.move_right_until_vertical,
             '7' : actions.move_one_up, '8' : actions.move_one_down, '9' : actions.move_one_left, 
             'a' : actions.move_one_right, 'b' : actions.move_up_until_horizontal, 'c' : actions.move_down_until_horizontal,
             'd' : actions.move_away_vertical, 'e' : actions.move_one_away_vertical, 'f': actions.move_away_horizontal,
             'g' : actions.move_one_away_horizontal}
        
        if dna is None:
            return
        else:
            self.populate(dna)
    
    def populate(self, dna):
        self.dna = dna
        genes = dna.split('0')
        if len(genes) != len(self.state_map):
            raise Exception('Invalid DNA: insufficiently many genes: ' + dna + ' (' + str(len(genes)) + '/' + str(len(self.state_map)) + ')')
        
        i = 1
        for gene in genes:
            ops = []
            for b in gene:
                op = self.funcs[b]
                ops.append(op)
            self.states[self.state_map[i]] = ops
            i += 1
    
    def get_ops(self, state):
        return self.states[state]
    
    def random(self):
        set = ['1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f', 'g']
        s = ''
        for i in range(len(self.states)):
            length = randint(1, 5)
            for j in range(length):
                s += set[randint(0, len(set) - 1)]
            s += '0'
        self.populate(s[:-1])
    
    def __str__(self):
        s = ''
        for state in self.states:
            s += state + ': ' + ' - '.join([f.func_name for f in self.states[state]]) + '\n'
        return s
    
    def state_description(self, state):
        return state + ': ' + ' - '.join([f.func_name for f in self.states[state]])

